<?php

/**
 * @file
 * Drupal stream wrapper implementation for Brandfolder.
 *
 * Implements DrupalStreamWrapperInterface to provide a Brandfolder wrapper
 * using the "bf://" scheme.
 */

/**
 * The stream wrapper class.
 */
class BrandfolderStreamWrapper implements DrupalStreamWrapperInterface {

  /**
   * Base URL.
   *
   * @var string
   */
  protected $baseUrl = NULL;

  /**
   * Brandfolder CDN Version.
   *
   * @var int
   */
  protected $cdnVersion = NULL;

  /**
   * Instance URI (stream).
   *
   * A stream is referenced as "scheme://target".
   *
   * @var string
   */
  protected $uri;

  /**
   * Filemime mapping.
   *
   * @var array
   *   Default map for determining filemime types.
   */
  protected static $mimeTypeMapping = NULL;

  /**
   * {@inheritdoc}
   *
   * @todo The following code doesn't seem to be necessary, but keep it on hand for now.
   * @code
   *   $this->context = stream_context_get_default();
   *   stream_context_set_option($this->context, 'bf', '<option>', '<value>');
   * @endcode
   *
   * @todo Consider performing common actions here (like loading a Brandfolder API client instance?) and storing results statically, a la:
   * @code
   *   $settings = &drupal_static('BrandfolderStreamWrapper_constructed_settings');
   *   if (!$settings) {
   *     $settings = [];
   *     $settings['client'] = brandfolder_api();
   *     // ...
   *   }
   * @endcode
   */
  public function __construct() {
    $this->cdnVersion = (int) variable_get('brandfolder_cdn_version', 1);
    switch ($this->cdnVersion) {
      case 2 :
        $this->baseUrl = 'https://cdn.bfldr.com';
        break;

      case 1 :
      default :
        $this->baseUrl = 'https://cdn.brandfolder.io';
        break;
    }
  }

  /**
   * Base implementation of setUri().
   */
  public function setUri($uri) {
    $this->uri = $uri;
  }

  /**
   * Base implementation of getUri().
   */
  public function getUri() {
    return $this->uri;
  }

  /**
   * Base implementation of realpath().
   */
  public function realpath() {
    return $this->getExternalUrl();
  }

  /**
   * Stream context resource.
   *
   * @var Resource
   */
  public $context;

  /**
   * A generic resource handle.
   *
   * @var Resource
   */
  public $handle = NULL;

  /**
   * Stat-related constant.
   *
   * As part of the inode protection mode returned by stat(), identifies the
   * file as a regular file, as opposed to a directory, symbolic link, or other
   * type of "file".
   *
   * @see http://linux.die.net/man/2/stat
   */
  const S_IFREG = 0100000;

  /**
   * Template for stat calls.
   *
   * All elements must be initialized.
   *
   * @var array
   */
  // @codingStandardsIgnoreStart
  protected $_stat = array(
    // @codingStandardsIgnoreEnd
    // Device number.
    0 => 0,
    'dev' => 0,
    // Inode number.
    1 => 0,
    'ino' => 0,
    // Inode protection mode. file_unmanaged_delete() requires is_file() to
    // return TRUE.
    2 => self::S_IFREG,
    'mode' => self::S_IFREG,
    // Number of links.
    3 => 0,
    'nlink' => 0,
    // Userid of owner.
    4 => 0,
    'uid' => 0,
    // Groupid of owner.
    5 => 0,
    'gid' => 0,
    // Device type, if inode device *.
    6 => -1,
    'rdev' => -1,
    // Size in bytes.
    7 => 0,
    'size' => 0,
    // Time of last access (Unix timestamp).
    8 => 0,
    'atime' => 0,
    // Time of last modification (Unix timestamp).
    9 => 0,
    'mtime' => 0,
    // Time of last inode change (Unix timestamp).
    10 => 0,
    'ctime' => 0,
    // Blocksize of filesystem IO.
    11 => -1,
    'blksize' => -1,
    // Number of blocks allocated.
    12 => -1,
    'blocks' => -1,
  );

  /**
   * Support for stat().
   */
  // @codingStandardsIgnoreStart
  public function url_stat($url, $flags) {
    // @codingStandardsIgnoreEnd
    // Load file size from DB before proceeding.
    $this->loadFileData($url);

    return $this->stream_stat();
  }

  /**
   * Determine the mime type for a given URI.
   */
  public static function getMimeType($uri, $mapping = NULL) {

    // Load the default mime type map.
    if (!isset(self::$mimeTypeMapping)) {
      include_once DRUPAL_ROOT . '/includes/file.mimetypes.inc';
      self::$mimeTypeMapping = file_mimetype_mapping();
    }

    // If a mapping wasn't specified, use the default map.
    if ($mapping == NULL) {
      $mapping = self::$mimeTypeMapping;
    }

    $extension = '';
    $file_parts = explode('.', basename($uri));

    // Remove the first part: a full filename should not match an extension.
    array_shift($file_parts);

    // Iterate over the file parts, trying to find a match.
    // For my.awesome.image.jpeg, we try:
    // - jpeg
    // - image.jpeg
    // - awesome.image.jpeg.
    while ($additional_part = array_pop($file_parts)) {
      $extension = strtolower($additional_part . ($extension ? '.' . $extension : ''));
      if (isset($mapping['extensions'][$extension])) {
        return $mapping['mimetypes'][$mapping['extensions'][$extension]];
      }
    }

    // No mime type matches, so return the default.
    return 'application/octet-stream';
  }

  /**
   * Support for fopen(), file_get_contents(), file_put_contents() etc.
   *
   * @param string $url
   *   A string containing the path to the file to open.
   * @param string $mode
   *   The file mode ("r", "wb" etc.).
   * @param bitmask $options
   *   A bit mask of STREAM_USE_PATH and STREAM_REPORT_ERRORS.
   * @param string &$opened_url
   *   A string containing the path actually opened.
   *
   * @return bool
   *   TRUE if file was opened successfully.
   */
  // @codingStandardsIgnoreStart
  public function stream_open($url, $mode, $options, &$opened_url) {
    // @codingStandardsIgnoreEnd

    // We only handle Read-Only mode by default.
    if ($mode != 'r' && $mode != 'rb') {
      return FALSE;
    }

    if ($options & STREAM_USE_PATH) {
      $opened_url = $url;
    }

    return TRUE;
  }

  /**
   * Get the absolute URL for a Brandfolder file.
   */
  public function getExternalUrl() {
    // The current approach is to store most of the URL information right in the
    // URI, so we do not need to perform any additional lookups against the
    // BF API or Drupal DB. Thus, a typical URI looks something like
    // "bf://SH123456/as/abc123-echvmo-7qf0za/my_image.jpg."
    // This is not quite as slick or readable as
    // "bf://abc123-echvmo-7qf0za/my_image.jpg"
    // or "bf://abc123-echvmo-7qf0za," but it should be more
    // performant and allow for straightforward management of (a) assets from
    // multiple Brandfolders and/or (b) files corresponding to different
    // attachments on the same BF asset if we choose to support that in the
    // future.
    // The effective limit on BF asset filenames (including extension) is
    // therefore 217 characters.
    $url_options = [
      'absolute' => TRUE,
    ];

    $scheme_prefix = 'bf://';
    $uri_sans_scheme = substr($this->getUri(), strlen($scheme_prefix));
    $query_params = [];

    // Handle image styles.
    $image_style = NULL;
    if (preg_match("/^styles\/([^\/]+)\/bf\/(.*)$/", $uri_sans_scheme, $matches)) {
      $image_style = $matches[1];
      // Remove the style portion of the URI.
      $uri_sans_scheme = $matches[2];
      $all_image_styles = image_styles();
      if (isset($all_image_styles[$image_style])) {
        // Apply all effects from the given image style. Our image toolkit will
        // handle compatible effects and add corresponding Smart CDN URL
        // transformation params to the image object.
        // @todo Test scenarios with stacked effects; try to provide more robust pass-through support for non BF images.
        $style = $all_image_styles[$image_style];
        if ($image = image_load("bf://$uri_sans_scheme")) {
          foreach ($style['effects'] as $effect) {
            image_effect_apply($image, $effect);
          }
          if (!empty($image->brandfolder['url_params'])) {
            // Safeguard against illegal values for crop zone and final image
            // size. These can occur when the image metadata is missing, etc.
            // The assumption here is that it is better to fall back to a
            // non-transformed image than to nothing.
            $bf_params =& $image->brandfolder['url_params'];
            if (isset($bf_params['crop_width']) && empty($bf_params['crop_width']) || isset($bf_params['crop_height']) && empty($bf_params['crop_height'])) {
              unset($bf_params['crop_width'], $bf_params['crop_height']);
            }
            if (isset($bf_params['width']) && empty($bf_params['width']) || isset($bf_params['height']) && empty($bf_params['height'])) {
              unset($bf_params['width'], $bf_params['height']);
            }
            $query_params = array_merge($query_params, $bf_params);
          }
        }
      }
    }

    // Lastly, convert the file format/extension to match the globally
    // configured preference if applicable. It's important to do this as late
    // as possible so other modules can accurately map the URI to a managed
    // file.
    $default_format = variable_get('brandfolder_default_cdn_static_image_file_format', FALSE);
    $extension_pattern = '/\.(\w+)(\?[^\?]*)?$/';
    $extension_found = preg_match($extension_pattern, $uri_sans_scheme, $matches);
    // Exempt GIF files from this conversion so as not to interfere with
    // animated GIF behavior.
    // @todo More sophisticated/granular handling for various file types.
    if ($default_format) {
      if ($extension_found) {
        $extension = strtolower($matches[1]);
        $convertible_extensions = [
          'jpg',
          'jpeg',
          'png',
          'webp',
          'auto',
        ];
        if (in_array($extension, $convertible_extensions) && $extension != $default_format) {
          $uri_sans_scheme = preg_replace($extension_pattern, ".$default_format$2", $uri_sans_scheme);
          $extension = $default_format;
        }
      }
      else {
        $uri_sans_scheme = preg_replace('/(\?[^\?]*)?$/', ".$default_format$1", $uri_sans_scheme);
        $extension = $default_format;
      }
    }

    $url = "{$this->baseUrl}/{$uri_sans_scheme}";
    // Remove any query params from the original URL and add them to the query
    // params array.
    $url_components = parse_url($url);
    if (!empty($url_components['query'])) {
      $original_query = $url_components['query'];
      $url = str_replace("?$original_query", '', $url);
      $original_query_pairs = explode('&', $original_query);
      foreach ($original_query_pairs as $pair) {
        $key_value = explode('=', $pair);
        if (count($key_value) == 2) {
          [$key, $value] = $key_value;
          $query_params[$key] = $value;
        }
      }
    }

    if ($this->cdnVersion == 2) {
      // @todo: Add support for the Fastly 'format' URL param in addition/as opposed to the file extension.
      $lossy_extensions = [
        'png8',
        'jpg',
        'pjpg',
        'webp',
        'webply',
      ];
      if (in_array($extension, $lossy_extensions)) {
        $img_quality = variable_get('brandfolder_cdn_default_image_quality', 85);
        $query_params['quality'] = $img_quality;
        // @todo: Also add the WEBP quality setting if the format is 'auto'.
      }
    }

    $url_options['query'] = $query_params;

    drupal_alter('brandfolder_file_external_url', $url, $url_options, $image_style);

    return url($url, $url_options);
  }

  /**
   * Get file data for a given URI.
   */
  protected function loadFileData($uri) {
    // @todo static/cache; data other than file size; etc.
    $result = db_select('brandfolder_file', 'bf')
      ->fields('bf', ['filesize'])
      ->condition('uri', $uri)
      ->execute();

    if ($result->rowCount() > 0) {
      $row = $result->fetch();
      if (!empty($row->filesize)) {
        // Set the appropriate items in the _stat array, which is used to
        // deliver data in response to file-system-esque requests.
        $this->_stat[7] = $this->_stat['size'] = $row->filesize;
      }
    }
  }

  /**
   * DrupalStreamWrapperInterface implementations, etc.
   */

  /**
   * Undocumented PHP stream wrapper method.
   */
  // @codingStandardsIgnoreStart
  public function stream_lock($operation) {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Support for fread(), file_get_contents() etc.
   *
   * @param int $count
   *   Maximum number of bytes to be read.
   *
   * @return bool
   *   The string that was read, or FALSE in case of an error.
   */
  // @codingStandardsIgnoreStart
  public function stream_read($count) {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Support for fwrite(), file_put_contents() etc.
   *
   * Since this is a read only stream wrapper this always returns false.
   *
   * @param string $data
   *   The string to be written.
   *
   * @return bool
   *   Returns FALSE.
   */
  // @codingStandardsIgnoreStart
  public function stream_write($data) {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Support for feof().
   *
   * @return bool
   *   TRUE if end-of-file has been reached.
   */
  // @codingStandardsIgnoreStart
  public function stream_eof() {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Support for fseek().
   *
   * @param int $offset
   *   The byte offset to got to.
   * @param string $whence
   *   SEEK_SET, SEEK_CUR, or SEEK_END.
   *
   * @return bool
   *   TRUE on success
   */
  // @codingStandardsIgnoreStart
  public function stream_seek($offset, $whence) {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Support for fflush().
   *
   * @return bool
   *   TRUE if data was successfully stored (or there was no data to store).
   */
  // @codingStandardsIgnoreStart
  public function stream_flush() {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Support for ftell().
   *
   * @return bool
   *   The current offset in bytes from the beginning of file.
   */
  // @codingStandardsIgnoreStart
  public function stream_tell() {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Support for fstat().
   *
   * @return array
   *   An array with file status, or FALSE in case of an error - see fstat()
   *   for a description of this array.
   */
  // @codingStandardsIgnoreStart
  public function stream_stat() {
    // @codingStandardsIgnoreEnd
    return $this->_stat;
  }

  /**
   * Support for fclose().
   *
   * @return bool
   *   TRUE if stream was successfully closed.
   */
  // @codingStandardsIgnoreStart
  public function stream_close() {
    // @codingStandardsIgnoreEnd
    return TRUE;
  }

  /**
   * Support for opendir().
   *
   * @param string $url
   *   A string containing the url to the directory to open.
   * @param int $options
   *   Whether or not to enforce safe_mode (0x04).
   *
   * @return bool
   *   TRUE on success.
   */
  // @codingStandardsIgnoreStart
  public function dir_opendir($url, $options) {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Support for readdir().
   *
   * @return bool
   *   The next filename, or FALSE if there are no more files in the directory.
   */
  // @codingStandardsIgnoreStart
  public function dir_readdir() {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Support for rewinddir().
   *
   * @return bool
   *   TRUE on success.
   */
  // @codingStandardsIgnoreStart
  public function dir_rewinddir() {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Support for closedir().
   *
   * @return bool
   *   TRUE on success.
   */
  // @codingStandardsIgnoreStart
  public function dir_closedir() {
    // @codingStandardsIgnoreEnd
    return FALSE;
  }

  /**
   * Gets the path that the wrapper is responsible for.
   *
   * This function isn't part of DrupalStreamWrapperInterface, but the rest
   * of Drupal calls it as if it were, so we need to define it.
   *
   * @return string
   *   The empty string. Since this is a remote stream wrapper,
   *   it has no directory path.
   */
  public function getDirectoryPath() {

    return '';
  }

  /**
   * Implements DrupalStreamWrapperInterface::unlink().
   */
  public function unlink($uri) {
    // Although the remote file itself can't be deleted, return TRUE so that
    // file_delete() can remove the file record from the Drupal database.
    return TRUE;
  }

  /**
   * Implements DrupalStreamWrapperInterface::rename().
   */
  public function rename($from_uri, $to_uri) {
    return FALSE;
  }

  /**
   * Implements DrupalStreamWrapperInterface::mkdir().
   */
  public function mkdir($uri, $mode, $options) {
    return FALSE;
  }

  /**
   * Implements DrupalStreamWrapperInterface::rmdir().
   */
  public function rmdir($uri, $options) {
    return FALSE;
  }

  /**
   * Implements DrupalStreamWrapperInterface::chmod().
   */
  public function chmod($mode) {
    return FALSE;
  }

  /**
   * Implements DrupalStreamWrapperInterface::dirname().
   */
  public function dirname($uri = NULL) {
    return FALSE;
  }

}
