<?php

/**
 * @file
 * Administration form for Brandfolder.
 */

use GuzzleHttp\Exception\GuzzleException;

/**
 * Builds the Settings form.
 */
function brandfolder_settings() {
  $form = array();

  $default_brandfolder = variable_get('brandfolder_default_brandfolder');
  $default_collection = variable_get('brandfolder_default_collection');
  $brandfolders_list = $collections_list = [];
  $api_key = variable_get('brandfolder_guest_api_key', FALSE);
  if ($api_key) {
    $bf = brandfolder_api($api_key);
    try {
      $brandfolders_list = $bf->getBrandfolders();
    }
    catch (Exception $e) {
      drupal_set_message(t('After you enter an API key, you can select a default Brandfolder'));
    }
    try {
      $collections_list = $default_brandfolder ? $bf->getCollectionsInBrandfolder($default_brandfolder) : [];
    }
    catch (Exception $e) {
      drupal_set_message(t('After you choose a default Brandfolder, you can select a default collection if you wish'));
    }
  }
  $none_options = [
    '_none' => t('<None>'),
  ];
  $collections_list = array_merge($none_options, $collections_list);

  /************************************
   * Credentials
   ************************************/
  $form['brandfolder_credentials'] = array(
    '#type'        => 'fieldset',
    '#title'       => t('Brandfolder Credentials'),
    '#collapsible' => TRUE,
    '#collapsed'   => variable_get('brandfolder_guest_api_key'),
  );

  // @todo Use separate Admin and/or Collaborator account API key if/when we add support for writing files from Drupal to Brandfolder (or generic ability to map Drupal roles to Brandfolder roles to customize access, etc.).
  $form['brandfolder_credentials']['brandfolder_guest_api_key'] = [
    '#type'          => 'textfield',
    '#title'         => t('Brandfolder API Key'),
    '#description'   => t('The API key from a Brandfolder user account that has access to the Brandfolder/collection where your images will be stored. For security reasons, it is recommended to create an account with the "Guest" role in Brandfolder and use the API key from that account.'),
    '#maxlength'     => 256,
    '#default_value' => variable_get('brandfolder_guest_api_key'),
  ];

  /************************************
   * Basic Configuration
   ************************************/
  // @todo which org/brandfolder/collection(s) to use, etc. Other media browser-related settings and Drupal role mapping?
  $form['basic'] = array(
    '#type'        => 'fieldset',
    '#title'       => t('Basic Configuration Options'),
    '#collapsible' => TRUE,
    '#collapsed'   => FALSE,
  );

  $form['basic']['brandfolder_default_brandfolder'] = [
    '#type'          => 'select',
    '#title'         => t('Default Brandfolder'),
    '#options'       => $brandfolders_list,
    '#default_value' => $default_brandfolder,
    '#description'   => t('The Brandfolder to use for all operations unless otherwise specified.'),
  ];

  $form['basic']['brandfolder_default_collection'] = [
    '#type'          => 'select',
    '#title'         => t('Default Collection'),
    '#options'       => $collections_list,
    '#default_value' => is_null($default_collection) ? '_none' : $default_collection,
    '#description'   => t('The collection to use for all operations unless otherwise specified.'),
  ];

  /************************************
   * Rich Text Editor Configuration
   ************************************/
  $form['rte'] = array(
    '#type'        => 'fieldset',
    '#title'       => t('WYSIWYG/Rich Text Editor Settings'),
    '#collapsible' => TRUE,
    '#collapsed'   => TRUE,
  );

  $form['rte']['brandfolder_browser_rte_button_label'] = [
    '#type'          => 'textfield',
    '#title'         => t('Brandfolder Browser Button Label'),
    '#description'   => t('The label for the button used to summon the Brandfolder Browser from a WYSIWYG editor toolbar (usually displayed on hover).'),
    '#maxlength'     => 64,
    '#default_value' => variable_get('brandfolder_browser_rte_button_label', t('Insert an Image')),
  ];

  $form['rte']['brandfolder_browser_rte_image_url_params'] = [
    '#type'          => 'textfield',
    '#title'         => t('Image Customizations'),
    '#description'   => t('(Optional) parameters that will be added to the CDN URL for any image embedded via WYSIWYG editor. You can use any string supported by the Brandfolder Smart CDN (e.g. to scale all images down to a width of 620 pixels, enter "width=620").'),
    '#maxlength'     => 1024,
    '#default_value' => variable_get('brandfolder_browser_rte_image_url_params', ''),
  ];


  /************************************
   * CDN Configuration
   ************************************/
  $form['cdn'] = array(
    '#type'        => 'fieldset',
    '#title'       => t('CDN Options'),
    '#collapsible' => TRUE,
    '#collapsed'   => TRUE,
  );

  $form['cdn']['brandfolder_default_cdn_static_image_file_format'] = [
    '#type'          => 'select',
    '#title'         => t('Default Static Image File Format'),
    // @codingStandardsIgnoreStart
    '#options'       => [
      '_none' => t('None'),
      'auto' => t('AUTO'),
      'jpg'  => 'JPG',
      'png'  => 'PNG',
      'webp' => 'WEBP',
    ],
    // @codingStandardsIgnoreEnd
    '#default_value' => variable_get('brandfolder_default_cdn_static_image_file_format', '_none'),
    '#description'   => t('If selected, this format will be used when generating CDN URLs for any assets whose native file format is PNG, JPG, or WEBP. The "AUTO" option causes a .webp file to be returned if the requesting browser supports that format, or a .png file otherwise.'),
  ];

  $form['cdn']['brandfolder_cdn_version'] = [
    '#type'          => 'select',
    '#title'         => t('CDN Version'),
    '#options'       => [
      1 => 'v1',
      2 => 'v2',
    ],
    '#default_value' => variable_get('brandfolder_cdn_version', 1),
    '#description'   => t('The version of the Brandfolder CDN enabled for your account. If you are unsure which version to use, contact Brandfolder support.'),
  ];

  if (variable_get('brandfolder_cdn_version', 1) == 2) {
    $form['cdn']['brandfolder_cdn_default_image_quality'] = [
      '#type'          => 'textfield',
      '#title'         => t('Default Image Quality'),
      '#description'   => t('A value between 1 and 100 controlling the image quality/fidelity for lossy image formats (e.g. JPG).'),
      '#maxlength'     => 3,
      '#default_value' => variable_get('brandfolder_cdn_default_image_quality', 85),
    ];

    $form['cdn']['brandfolder_cdn_default_image_quality_webp'] = [
      '#type'          => 'textfield',
      '#title'         => t('Default Image Quality for WEBP'),
      '#description'   => t('A value between 1 and 100 controlling the image quality/fidelity for WEBP images when using auto=webp mode.'),
      '#maxlength'     => 3,
      '#default_value' => variable_get('brandfolder_cdn_default_image_quality_webp', NULL),
    ];

    $form['cdn']['brandfolder_cdn_safe_cropping_mode'] = [
      '#type'          => 'checkbox',
      '#title'         => t('Use safe cropping mode'),
      '#description'   => t('Use the "safe" option when specifying crop dimensions via the image optimization API. This will avoid "Invalid crop, region out of bounds" errors, but may result in distorted images when crop values are not deemed compatible with the source image. See https://developer.fastly.com/reference/io/crop.'),
      '#default_value' => variable_get('brandfolder_cdn_safe_cropping_mode', 1),
    ];
  }

  /************************************
   * Webhooks
   ************************************/
  $form['webhooks'] = array(
    '#type'        => 'fieldset',
    '#title'       => t('Webhooks'),
    '#collapsible' => TRUE,
    '#collapsed'   => TRUE,
  );

  $form['webhooks']['brandfolder_webhooks_debug'] = [
    '#type'          => 'checkbox',
    '#title'         => t('Log all webhook events (e.g. for debugging purposes)'),
    '#default_value' => variable_get('brandfolder_webhooks_debug', FALSE),
  ];

  /************************************
   * Maintenance/Sync
   ************************************/
  $form['sync'] = array(
    '#type'        => 'fieldset',
    '#title'       => t('Maintenance/Sync Settings'),
    '#collapsible' => TRUE,
    '#collapsed'   => TRUE,
  );

  $form['sync']['brandfolder_cron_asset_deprecation_batch_size'] = [
    '#type'          => 'textfield',
    '#title'         => t('Asset deprecation check: Batch size'),
    '#description'   => t('Whenever the brandfolder_cron() function runs, it will look for any Drupal files whose Brandfolder source is no longer accessible, and notify other Drupal modules of any such files/assets. Use the system-wide cron settings or the Elysia Cron module to control when and how frequently this process occurs. Change this setting here to control how many Drupal-Brandfolder files are processed each time it runs.'),
    '#maxlength'     => 12,
    '#default_value' => variable_get('brandfolder_cron_asset_deprecation_batch_size', 250),
  ];

  $form = system_settings_form($form);

  if (isset($form['#validate']) && is_array($form['#validate'])) {
    array_unshift($form['#validate'], 'brandfolder_settings_form_validate');
  }
  else {
    $form['#validate'] = [
      'brandfolder_settings_form_validate',
    ];
  }

  // Display some images from the selected Brandfolder/collection if applicable.
  if ($default_brandfolder) {
    $bf->default_brandfolder_id = $default_brandfolder;
    if ($default_collection) {
      $assets = $bf->listAssets([], $default_collection);
    }
    else {
      $assets = $bf->listAssets();
    }
    if ($assets) {
      $thumbnails = array_map(function ($asset) {
        return '<img src="' . $asset->attributes->thumbnail_url . '">';
      }, $assets->data);

      $form['sample_pics'] = [
        '#type' => 'markup',
        '#prefix' => '<h2>Sample Images</h2>',
        '#markup' => implode(' ', $thumbnails),
      ];
    }
  }

  return $form;
}

/**
 * Form validation handler for settings form.
 *
 * @param array $form
 *   The Drupal form array.
 * @param array $form_state
 *   The Drupal form state array containing values, etc.
 */
function brandfolder_settings_form_validate(array &$form, array &$form_state) {
  $api_key = $form_state['values']['brandfolder_guest_api_key'];
  $api_success = FALSE;
  if (!empty($api_key)) {
    $bf = brandfolder_api($api_key);
    try {
      $brandfolders = $bf->getBrandfolders();
      // Note that the getBrandfolders request will return a 200 response even
      // if the API key is invalid, and the brandfolders array will simply be
      // empty. This is a quirk of the Brandfolder API.
      if (!empty($brandfolders)) {
        $api_success = TRUE;
      }
    }
    catch (GuzzleException $e) {
      $api_success = FALSE;
    }
    if (!$api_success) {
      $message = t('Could not connect to Brandfolder using this API key. Make sure the key is correct and is linked to a Brandfolder user who has permission to access at least one Brandfolder.');
      form_set_error('brandfolder_guest_api_key', $message);
    }
  }
  if ($api_success) {
    $old_brandfolder = variable_get('brandfolder_default_brandfolder');
    $specified_brandfolder = $form_state['values']['brandfolder_default_brandfolder'];
    // If the Brandfolder selection is being changed, reset the collection,
    // which is dependent on the Brandfolder. Otherwise, use the value
    // specified by the form (while also ensuring that a _none selection is
    // stored as NULL).
    $specified_collection = $form_state['values']['brandfolder_default_collection'];
    if ($specified_brandfolder != $old_brandfolder || $specified_collection == '_none') {
      $form_state['values']['brandfolder_default_collection'] = NULL;
    }
  }
  else {
    // If no (good) API key is specified, clear out any existing Brandfolder and
    // Collection choices.
    $form_state['values']['brandfolder_default_brandfolder'] = NULL;
    $form_state['values']['brandfolder_default_collection'] = NULL;
  }
  if ($form_state['values']['brandfolder_default_cdn_static_image_file_format'] == '_none') {
    $form_state['values']['brandfolder_default_cdn_static_image_file_format'] = NULL;
  }
}
